//Title: Graphics Device Class
//Platform: BeOS 5
//Version: 2001.02.03
//Description:
//	A class for working with physical graphics devices (like VGA displays)
//  Implements a high-level interface to the BeOS graphics drivers and accelerants

//Copyright (C) 2002 Kevin H. Patterson

#ifndef _GraphicsDevice_h
#define _GraphicsDevice_h

	#include <Application.h>

	//For 'MediaNode' code
		//#include <Area.h>
		#include <Buffer.h>
		#include <BufferGroup.h>

	//For 'BeTVOut' code
		#define VERSION 117	//app version number = 0.11.7
		
		#include "BeTVOut/riva_iic/riva_iic.h"
		#include "BeTVOut/BT868_9.h"
		#include "BeTVOut/Transfer.h"
		#include "BeTVOut/Chrontel.h"

	//For 'slimdemo' code
		#include <OS.h>
		#include <image.h>
		#include <FindDirectory.h>
		#include <graphic_driver.h>
		#include <Accelerant.h>
		#include <dirent.h>
		#include <stdio.h>
		#include <string.h>
		#include <sys/stat.h>
		#include <malloc.h>
		#include <errno.h>
		#include <stdlib.h>
	
	class GraphicsDevice
	{
		friend class DeviceManager;
	private:
		GraphicsDevice(const char * DeviceName);
		bool InitCheck();

	public:
		~GraphicsDevice();
		const char * DeviceName();
		const char * Name();
		const char * Chipset();
		const char * SerialNumber();
		uint32 Memory();
		uint32 DACSpeed();
		
		BBufferGroup * CreateBufferGroup();
		void DeleteBufferGroup(BBufferGroup * uBufferGroup);
		bool SetTVOutMode(const int Mode, const int OutputMode = 3, const bool FilckerFiltering = true, const bool DualDisplay = true);

	private:
		image_id LoadAccelerantImage();
		int nonVESAsetVESA640(display_mode new_mode);
		int nonVESAsetVESA800(display_mode new_mode);
		status_t ScreenSetMode(display_mode * new_mode, bool);
		
	private:
		const char * iDeviceName;
		bool iInitCheck;
		int hDevice;
		image_id AccelerantImage;
		GetAccelerantHook GetAccelerantFunction;
		accelerant_device_info * AccelerantDeviceInfo;

	};
	
#endif
