// VideoConsumer.cpp

#ifndef _VideoConsumer_h
#define _VideoConsumer_h
	
	#include <MediaNode.h>
	#include <MediaEventLooper.h>
	#include <BufferConsumer.h>
	#include <Controllable.h>
	#include <TimedEventQueue.h>
	#include <OS.h>
	#include <String.h>
	
	#include "DeviceManager.h"
	#include "GraphicsDevice.h"
	
	class VideoConsumer
	:
		public BMediaEventLooper,
		public BBufferConsumer,
		public BControllable
	{
	public:
		explicit VideoConsumer(
		BMediaAddOn * addon,
		const char * name,
		int32 internalID );
		
		virtual ~VideoConsumer();
		
		status_t InitCheck() const;
		
		status_t GetConfiguration( BMessage * outMessage ) const;
		
		status_t SetConfiguration( BMessage * inMessage );
		
	/* BMediaNode */
	public:
		port_id ControlPort() const;
		
		BMediaAddOn * AddOn( int32 * cookie ) const;
		
		status_t HandleMessage(
			int32 message,
			const void * data,
			size_t size );
		
	protected:
		status_t RequestCompleted( const media_request_info & info );
		
	/* BMediaEventLooper */
	protected:
		void NodeRegistered();
		
		void Start( bigtime_t performance_time );
		
		void Stop( bigtime_t performance_time, bool immediate );
		
		void Seek( bigtime_t media_time, bigtime_t performance_time );
		
		void TimeWarp( bigtime_t at_real_time, bigtime_t to_performance_time );
		
		void HandleEvent(
			const media_timed_event * event,
			bigtime_t lateness,
			bool realTimeEvent );
		
		status_t DeleteHook( BMediaNode * node );
		
	/* BBufferConsumer */
	public:
		status_t AcceptFormat(
			const media_destination & dest,
			media_format * format );
		
		status_t GetNextInput(
			int32 * cookie,
			media_input * outInput );
		
	protected:
		void DisposeInputCookie( int32 cookie );
		
		void BufferReceived( BBuffer * buffer );
		
		void ProducerDataStatus(
			const media_destination &for_whom,
			int32 status,
			bigtime_t at_media_time );
		
		status_t GetLatencyFor(
			const media_destination &for_whom,
			bigtime_t * outLatency,
			media_node_id * out_id );
		
		status_t Connected(
			const media_source &producer,
			const media_destination &where,
			const media_format & with_format,
			media_input * outInput );
		
		void Disconnected(
			const media_source &producer,
			const media_destination &where );
		
		status_t FormatChanged(
			const media_source & producer,
			const media_destination & consumer,
			int32 from_change_count,
			const media_format & format );
		
	/* BControllable */
	protected:
		status_t GetParameterValue(
			int32 id,
			bigtime_t * last_change,
			void * value,
			size_t * size );
		
		void SetParameterValue(
			int32 id,
			bigtime_t when,
			const void * value,
			size_t size );
		
		status_t StartControlPanel( BMessenger * messenger );
		
	/* Implementation */
	private:
		status_t SpecializeFormat( media_format * format );
		
		void StartVideoOutput();
		
		void StopVideoOutput();
		
		void UpdateVideoParameter( int32 id );
		
		BParameterWeb * CreateParameterWeb();
		
	private:
		static vint32 fInstances;
		
		status_t fStatus;
		
		bool fPassthru;
		
		BMediaAddOn * fAddOn;
		int32 fInternalID;
		
		BBufferGroup * fBufferGroup;
		uint32 fBufferOffset;
		uint32 fBufferSize;
		
		bigtime_t fLatency;
		bigtime_t fProcessingLatency;
		
		media_input fInput;
		media_format fPreferredFormat;
		
		BWindow * fWindow;
		
	private:
		enum {
			Param_DeviceID,
			Param_DisplayMode,
			Param_VideoFormat,
			Param_VideoImage,
			Param_VideoOutput,
			Param_VideoFiltering
		};
		
		int32 fDeviceID;
		int32 fDisplayMode;
		int32 fVideoFormat;
		int32 fVideoImage;
		int32 fVideoOutput;
		float fVideoFiltering;
		
		bigtime_t fDeviceIDTime;
		bigtime_t fDisplayModeTime;
		bigtime_t fVideoFormatTime;
		bigtime_t fVideoImageTime;
		bigtime_t fVideoOutputTime;
		bigtime_t fVideoFilteringTime;
		
		DeviceManager * tDeviceManager;
		GraphicsDevice * tGraphicsDevice;
	};
	
#endif
